"""SQL Error Codes"""

# SQL compilation error: Object ''{0}'' does not exist or not authorized.
OBJECT_NOT_EXIST = 2001
# SQL compilation error: Object ''{0}'' already exists.
OBJECT_ALREADY_EXISTS = 2002
